from litellm.llms.openai.image_generation import DallE3ImageGenerationConfig


class AzureDallE3ImageGenerationConfig(DallE3ImageGenerationConfig):
    """
    Azure dall-e-3 image generation config
    """

    pass
